#include "aio.h"
#include "buffer_manager.h"
#include "config.h"
#include "ds.h"
#include "file_tab.h"
#include "fsutils.h"
#include "hashing.h"
#include "renderer_api.h"
#include "resources.h"
#include "threading.h"

#include <assert.h>
#include <stdlib.h>
#include <string.h>

RT_CVAR_S(rt_ResourceDirectory, "The directory used for storing resources. Default: res", "res");
RT_CVAR_I(rt_ResourceCacheSize,
          "The maximum amount of memory used for caching resources. Default: 512MB",
          RT_MB(512));
RT_CVAR_I(rt_MaxCachedResources,
          "The maximum number of simultaneously cached resources. Default: 1024",
          1024);
RT_CVAR_I(rt_ResourceNamespaceSize,
          "The maximum number of resources that can exist. Default: 1.048.576",
          1048576);

#define RT_TOMBSTONE_ID 1

typedef struct {
    void *buffer;
    size_t size;
    int next_free;
    int usage_counter;
} rt_cached_resource;

typedef struct {
    unsigned int index;
    rt_resource_id id;
} rt_cached_resource_ref;

typedef struct {
    void *mem;

    rt_minheap reclaim_heap;

    /* Used to lookup cached resources by id */
    rt_resource_id *resource_ids;
    unsigned int *resource_indices;

    rt_cached_resource *resources;
    unsigned int first_free;

    size_t current_size;

    rt_rwlock lock;
} rt_resource_cache;

typedef struct {
    rt_file_id file;
    size_t offset;
    size_t size;
} rt_resource_ref;

typedef struct {
    rt_resource_id *ids;
    rt_resource_ref *refs;
    rt_rwlock lock;
} rt_resource_namespace;

/* ~~~ Utilities ~~~ */

static size_t GetResourceDataSize(const rt_resource *resource) {
    switch (resource->type) {
    case RT_RESOURCE_PIPELINE:
        return sizeof(rt_pipeline_info);
    case RT_RESOURCE_SHADER: {
        /* Sizeof metadata + bytecode */
        const rt_shader_info *info = resource->data;
        return sizeof(rt_shader_info) + (info) ? info->bytecode_length : 0;
    } break;
    default:
        rtLog("RESMGR", "Tried to get size of an invalid resource type %u", resource->type);
    }
    return 0;
}

static void CopyResourceData(const rt_resource *resource, void *dest) {
    switch (resource->type) {
    case RT_RESOURCE_PIPELINE:
        memcpy(dest, resource->data, sizeof(rt_pipeline_info));
        break;
    case RT_RESOURCE_SHADER: {
        /* Sizeof metadata + bytecode */
        const rt_shader_info *info = resource->data;
        rt_shader_info *dest_info  = dest;
        memcpy(dest_info, info, sizeof(*info));
        memcpy(dest_info + 1, rtResolveConstRelptr(&info->bytecode), info->bytecode_length);
        rtSetRelptr(&dest_info->bytecode, (void *)(dest_info + 1));
    } break;
    default:
        rtLog("RESMGR", "Tried to get copy a resource of invalid type %u", resource->type);
    }
}

/* ~~~ Cache ~~~ */

static rt_resource_cache _cache;

static rt_result InitResourceCache(void) {
    int count = rt_MaxCachedResources.i;
    if (count == 0) {
        rtReportError("RESMGR", "rt_MaxCachedResources must be greater than 0.");
        return RT_INVALID_VALUE;
    }

    size_t required_mem = (size_t)count * (sizeof(rt_cached_resource_ref) + sizeof(int) +
                                           sizeof(rt_cached_resource)) +
                          2 * (size_t)count * (sizeof(rt_resource_id) + sizeof(unsigned int));

    void *mem = malloc(required_mem);
    if (!mem)
        return RT_OUT_OF_MEMORY;
    rt_create_rwlock_result lock_create = rtCreateRWLock();
    if (!lock_create.ok) {
        free(mem);
        return RT_UNKNOWN_ERROR;
    }
    memset(mem, 0, required_mem);
    _cache.mem = mem;

    int *reclaim_keys                    = mem;
    rt_cached_resource_ref *reclaim_refs = (rt_cached_resource_ref *)reclaim_keys + count;
    _cache.reclaim_heap                  = rtCreateMinheap(reclaim_keys,
                                          reclaim_refs,
                                          sizeof(rt_cached_resource_ref),
                                          (size_t)count,
                                          0);

    _cache.current_size = 0;
    _cache.resources    = (rt_cached_resource *)(reclaim_keys + count);
    _cache.lock         = lock_create.lock;

    for (int i = 0; i < count; ++i) {
        _cache.resources[i].next_free = (i < count - 1) ? i + 1 : UINT_MAX;
    }
    _cache.first_free = 0;

    _cache.resource_ids     = (rt_resource_id *)(_cache.resources + count);
    _cache.resource_indices = (unsigned int *)(_cache.resource_ids + 2 * count);

    return RT_SUCCESS;
}

static void ShutdownResourceCache(void) {
    free(_cache.mem);
    rtDestroyRWLock(&_cache.lock);
    memset(&_cache, 0, sizeof(_cache));
}

static bool FreeCacheSpace(size_t space) {
    size_t total_freed = 0;
    while (total_freed < space && !rtMinheapIsEmpty(&_cache.reclaim_heap)) {
        rt_cached_resource_ref ref;
        rtMinheapPop(&_cache.reclaim_heap, &ref);

        rt_cached_resource *res = &_cache.resources[ref.index];
        rtReleaseBuffer(res->buffer, res->size);
        total_freed += res->size;

        res->next_free     = _cache.first_free;
        _cache.first_free  = ref.index;
        res->usage_counter = 0;
        res->buffer        = NULL;
        res->size          = 0;

        /* Remove from lookup table */
        size_t ht_size = (size_t)rt_MaxCachedResources.i * 2;
        for (size_t off = 0; off < ht_size; ++off) {
            size_t slot = (ref.id + off) % ht_size;
            if (_cache.resource_ids[slot] == ref.id) {
                _cache.resource_ids[slot] = RT_TOMBSTONE_ID;
                break;
            } else if (_cache.resource_ids[slot] == RT_INVALID_RESOURCE_ID) {
                break;
            }
        }
    }
    return total_freed >= space;
}

static unsigned int FindCachedResource(rt_resource_id id) {
    size_t ht_size = (size_t)rt_MaxCachedResources.i * 2;
    for (size_t off = 0; off < ht_size; ++off) {
        size_t slot = (id + off) % ht_size;
        if (_cache.resource_ids[slot] == id)
            return _cache.resource_indices[slot];
        else if (_cache.resource_ids[slot] == RT_INVALID_RESOURCE_ID)
            return UINT_MAX;
    }
    return UINT_MAX;
}

static rt_resource *CacheResource(rt_resource_id id, const rt_resource *res) {
    rt_resource *cached = NULL;
    rtLockWrite(&_cache.lock);
    unsigned int index = FindCachedResource(id);
    if (index != UINT_MAX) {
        rt_cached_resource_ref ref      = {.id = id, .index = index};
        rt_cached_resource *cache_entry = &_cache.resources[index];
        ++cache_entry->usage_counter;
        rtMinheapUpdate(&_cache.reclaim_heap, &ref, cache_entry->usage_counter, NULL);
        cached = cache_entry->buffer;
    } else {
        /* Insert into cache */
        size_t total_size = sizeof(rt_resource) + GetResourceDataSize(res);
        if (_cache.current_size + total_size >= (size_t)rt_ResourceCacheSize.i) {
            if (!FreeCacheSpace(total_size)) {
                rtLog("RESMGR",
                      "Unable to reclaim %zu kB from the resource cache.",
                      total_size / 1024);
                rtUnlockWrite(&_cache.lock);
                return NULL;
            }
            RT_ASSERT(_cache.first_free != UINT_MAX,
                      "There must be a free cache entry after space was freed.");
        }

        void *buffer = rtAllocBuffer(total_size);
        if (!buffer) {
            rtLog("RESMG", "Unable to allocate %zu kB for the new resource.", total_size / 1024);
            rtUnlockWrite(&_cache.lock);
            return NULL;
        }
        memcpy(buffer, res, sizeof(rt_resource));
        cached       = buffer;
        cached->data = (void *)(cached + 1);
        CopyResourceData(res, cached->data);

        index                                 = _cache.first_free;
        _cache.first_free                     = _cache.resources[index].next_free;
        _cache.resources[index].buffer        = buffer;
        _cache.resources[index].usage_counter = 1;
        _cache.resources[index].size          = total_size;
        _cache.resources[index].next_free     = UINT_MAX;

        rt_cached_resource_ref reclaim_ref = {
            .id    = id,
            .index = index,
        };
        rtMinheapPush(&_cache.reclaim_heap, 1, &reclaim_ref);

        /* Insert into lookup table */
        bool inserted  = false;
        size_t ht_size = (size_t)rt_MaxCachedResources.i * 2;
        for (size_t off = 0; off < ht_size; ++off) {
            size_t slot = (id + off) % ht_size;
            if (_cache.resource_ids[slot] == RT_INVALID_RESOURCE_ID ||
                _cache.resource_ids[slot] == RT_TOMBSTONE_ID || _cache.resource_ids[slot] == id) {
                _cache.resource_indices[slot] = index;
                inserted                      = true;
                break;
            }
        }
        if (!inserted) {
            rtReportError("RESMGR",
                          "Failed to insert created resource into the resource lookup table.");
        }
    }
    rtUnlockWrite(&_cache.lock);
    return cached;
}

/* ~~~ Resource Namespace ~~~ */

static rt_resource_namespace _namespace;

static rt_result InitResourceNamespace(void) {
    size_t size = (size_t)rt_ResourceNamespaceSize.i;
    if (size == 0) {
        rtReportError("RESMGR", "rt_ResourceNamespaceSize must be greater than 0.");
        return RT_INVALID_VALUE;
    }

    void *mem = calloc(size, sizeof(rt_resource_id) + sizeof(rt_resource_ref));
    if (!mem)
        return RT_OUT_OF_MEMORY;
    rt_create_rwlock_result lock_create = rtCreateRWLock();
    if (!lock_create.ok) {
        free(mem);
        return RT_UNKNOWN_ERROR;
    }
    _namespace.lock = lock_create.lock;
    _namespace.ids  = mem;
    _namespace.refs = (rt_resource_ref *)(_namespace.ids + size);
    return RT_SUCCESS;
}

static void ShutdownNamespace(void) {
    rtDestroyRWLock(&_namespace.lock);
    free(_namespace.ids);
    memset(&_namespace, 0, sizeof(_namespace));
}

#if 0
static rt_resource_ref *GetResourceRefPtr(rt_resource_id id) {
    rt_resource_ref *ref = NULL;
    rtLockRead(&_namespace.lock);
    size_t ns_size = (size_t)rt_ResourceNamespaceSize.i;
    for (size_t j = 0; j < ns_size; ++j) {
        size_t at = (id + j) % ns_size;
        if (_namespace.ids[at] == RT_INVALID_RESOURCE_ID) {
            break;
        } else if (_namespace.ids[at] == id) {
            ref = &_namespace.refs[at];
            break;
        }
    }
    rtUnlockRead(&_namespace.lock);
    return ref;
}
#endif

/* Fills the passed write struct with the necessary information to save the resource to a file */
static void
PrepareResourceFlushToFile(rt_resource_id id, const rt_resource *resource, rt_file_write *write) {
    /* A file write needs one contiguous buffer */
    RT_ASSERT(((uintptr_t)resource->data == (uintptr_t)resource + sizeof(*resource)),
              "The resource must reside in the cache, to ensure the correct memory layout");

    char file_path[260];
    rtSPrint(file_path, 260, "%s/%llx.bin", rt_ResourceDirectory.s, id);

    write->file      = rtAddFile(file_path);
    write->buffer    = resource;
    write->offset    = 0;
    write->num_bytes = sizeof(rt_resource) + GetResourceDataSize(resource);
}

/* ~~~ Public API ~~~ */

rt_result InitResourceManager(void) {
    if (!rtCreateDirectory(rt_ResourceDirectory.s))
        rtLog("RESMGR", "CreateDirectory(%s) failed.", rt_ResourceDirectory.s);

    rt_result res;
    if ((res = InitResourceCache()) != RT_SUCCESS)
        return res;
    if ((res = InitResourceNamespace()) != RT_SUCCESS) {
        ShutdownResourceCache();
        return res;
    }
    return RT_SUCCESS;
}

void ShutdownResourceManager(void) {
    ShutdownResourceCache();
    ShutdownNamespace();
}

RT_DLLEXPORT rt_result rtCreateResources(uint32_t count,
                                         const char **names,
                                         const rt_resource *resources,
                                         rt_resource_id *ids) {
    rt_result result = RT_SUCCESS;
    size_t ns_size   = (size_t)rt_ResourceNamespaceSize.i;

    rt_write_batch writes = {.num_writes = 0};
    rt_aio_handle write_handles[RT_WRITE_BATCH_MAX_SIZE];
    uint32_t outstanding_writes = 0;

    rtLockWrite(&_namespace.lock);
    for (uint32_t i = 0; i < count; ++i) {
        size_t name_len   = strlen(names[i]);
        rt_resource_id id = (rt_resource_id)rtHashBytes(names[i], name_len);
        if (id == RT_INVALID_RESOURCE_ID || id == RT_TOMBSTONE_ID)
            id = ~id;

        bool inserted = false;
        for (size_t j = 0; j < ns_size; ++j) {
            size_t at = (id + j) % ns_size;
            if (_namespace.ids[at] == RT_INVALID_RESOURCE_ID) {
                inserted = true;

                ids[i] = id;

                const rt_resource *cached_resource = CacheResource(id, &resources[i]);

                PrepareResourceFlushToFile(id, cached_resource, &writes.writes[writes.num_writes]);
                _namespace.ids[at]         = id;
                _namespace.refs[at].offset = writes.writes[writes.num_writes].offset;
                _namespace.refs[at].size   = writes.writes[writes.num_writes].num_bytes;
                _namespace.refs[at].file   = writes.writes[writes.num_writes].file;
                ++writes.num_writes;
                break;
            } else if (_namespace.ids[at] == id) {
                rtReportError("RESMGR",
                              "Resource ID collision occured with resource %s.\nID: %llx",
                              names[i],
                              id);
                result = RT_INVALID_FILE_ID;
                goto out;
            }
        }
        if (!inserted) {
            result = RT_OUT_OF_MEMORY;
            goto out;
        }

        if (writes.num_writes == RT_WRITE_BATCH_MAX_SIZE ||
            (i == count - 1 && writes.num_writes > 0)) {
            if (outstanding_writes > 0) {
                /* Wait until the previous batch is finished */
                for (uint32_t k = 0; k < outstanding_writes; ++k) {
                    if (rtWaitForAIOCompletion(write_handles[k]) != RT_AIO_STATE_FINISHED) {
                        rtReportError("RESMGR", "Resource write failed.");
                        result = RT_UNKNOWN_ERROR;
                        goto out;
                    }
                    rtReleaseAIO(write_handles[k]);
                }
            }
            outstanding_writes = writes.num_writes;
            if (rtSubmitWriteBatch(&writes, write_handles) != RT_SUCCESS) {
                rtReportError("RESMGR", "Failed to submit resource writes.");
                result = RT_UNKNOWN_ERROR;
                goto out;
            }
        }
    }

    if (outstanding_writes > 0) {
        /* Wait until the last batch is finished */
        for (uint32_t i = 0; i < outstanding_writes; ++i) {
            if (rtWaitForAIOCompletion(write_handles[i]) != RT_AIO_STATE_FINISHED) {
                rtReportError("RESMGR", "Resource write failed.");
                result = RT_UNKNOWN_ERROR;
            }
            rtReleaseAIO(write_handles[i]);
        }
    }
out:
    rtUnlockWrite(&_namespace.lock);
    return result;
}