#ifndef RT_RESOURCES_H
#define RT_RESOURCES_H

/* Resource system interface
 *
 * To differentiate the two ideas, we called processed assets "resources"
 * and the source files "assets".
 *
 * For example a .pipeline file is an asset, while a compiled pipeline in
 * a binary file is a resource.
 *
 * Furthermore, a single asset file might contain multiple resources,
 * i.e. a single texture file might be turned into multiple resources for the different mip-levels.
 */

#include <stdint.h>

#include "runtime.h"

#ifdef __cplusplus
extern "C" {
#endif

/* Identifies a single resource
 *
 * This is a hash of the resource name. */
typedef uint64_t rt_resource_id;

#define RT_INVALID_RESOURCE_ID 0u

typedef enum {
    /* Compiled shader code */
    RT_RESOURCE_SHADER,

    /* A pipeline state object */
    RT_RESOURCE_PIPELINE,
} rt_resource_type;

#define RT_MAX_SUBRESOURCES          32
#define RT_MAX_RESOURCE_DEPENDENCIES 32

typedef struct {
    /* Points to the resource data. The size of which is determined by the type. */
    void *data;
    rt_resource_type type;

    /* Subresources are necessary to complete the resource.
     * For example, a texture might contain different mip-levels as sub-resources. */
    uint32_t subresource_count;
    rt_resource_id subresources[RT_MAX_SUBRESOURCES];

    /* Dependencies reference distinct resources that are necessary to use this resource.
     * For example, a model file might depend on its materials */
    uint32_t dependency_count;
    rt_resource_id dependencies[RT_MAX_RESOURCE_DEPENDENCIES];
} rt_resource;

/* Registers resources with the resource manager, making them available to the system.
 * 
 * The runtime will create a standalone file for each resource in the resource directory.
 * To package them, you will need to use a separate tool.
 */
RT_DLLEXPORT rt_result rtCreateResources(uint32_t count,
                                         const char **names,
                                         const rt_resource *resources,
                                         rt_resource_id *ids);

#ifdef __cplusplus
}
#endif

#endif
