#include "dependency_tracking.h"

#define RT_DEFINE_DEPENDENCY_FILE_STRUCTURES
#include "runtime/assets.h"
#include "runtime/threading.h"
#include "runtime/asset_dependencies.h"

#include <stdbool.h>
#include <limits.h>
#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>


/* Track a list of dependencies per asset.
 * For the runtime, we only care about dependent -> dependency (so for example material -> texture),
 * but for assetc we also care about the opposite direction:
 * If an asset has the reprocess_on_dependency_change setting enabled,
 * we need to queue a reprocessing, even if the asset itself did not change.
 */

#define END_OF_LIST 0

/* 64 byte cache line - 8 (next index + count) / 4 (u32) = 14 */
#define BUCKET_ENTRY_COUNT 14
typedef struct rt_dep_list_bucket_s {
    uint32_t next;
    uint32_t count;
    rt_uid entries[BUCKET_ENTRY_COUNT];
} rt_dep_list_bucket;

typedef union {
    /* Indices of the first buckets */
    struct {
        uint32_t dependencies;
        uint32_t dependents;
    };
    uint32_t lists[2];
} rt_dep_list;

static rt_mutex *_guard;

static rt_dep_list_bucket *_buckets;
static uint32_t _bucket_count;
static uint32_t _bucket_capacity;


#define MAP_SIZE 2048
static rt_uid _uids[MAP_SIZE];
static rt_dep_list _lists[MAP_SIZE];

rt_result rtInitDependencyTracking(void) {
    _guard = rtCreateMutex();
    if (!_guard)
        return RT_UNKNOWN_ERROR;

    _buckets = malloc(sizeof(rt_dep_list_bucket) * 256);
    if (!_buckets)
        return RT_UNKNOWN_ERROR;
    _bucket_capacity = 256;
    _bucket_count    = 1;

    return RT_SUCCESS;
}

static uint32_t AllocNewBucket(void) {
    if (_bucket_count == _bucket_capacity) {
        void *t = realloc(_buckets, (size_t)_bucket_capacity * 2 * sizeof(rt_dep_list_bucket));
        if (!t)
            return 0;
        _buckets = t;
        _bucket_capacity *= 2;
    }
    memset(&_buckets[_bucket_count], 0, sizeof(rt_dep_list_bucket));
    return _bucket_count++;
}

static rt_result InsertIntoList(rt_uid list_asset, rt_uid uid, int list_index) {
    rtLockMutex(_guard);
    bool inserted = false;
    for (uint32_t i = 0; i < MAP_SIZE; i++) {
        uint32_t at = (list_asset + i) % MAP_SIZE;
        if (_uids[at] == list_asset || _uids[at] == 0) {
            _uids[at] = list_asset;

            /* Alloc a new list, if necessary */
            if (_lists[at].lists[list_index] == 0) {
                _lists[at].lists[list_index] = AllocNewBucket();
                if (!_lists[at].lists[list_index]) {
                    rtUnlockMutex(_guard);
                    return RT_UNKNOWN_ERROR;
                }
            }

            /* Advance to the end of the list */
            rt_dep_list_bucket *bucket = &_buckets[_lists[at].lists[list_index]];
            while (bucket->next != END_OF_LIST) {
                bucket = &_buckets[bucket->next];
            }

            /* Grow the list, if necessary */
            if (bucket->count == BUCKET_ENTRY_COUNT) {
                bucket->next = AllocNewBucket();
                if (!bucket->next) {
                    rtUnlockMutex(_guard);
                    return RT_UNKNOWN_ERROR;
                }
                bucket = &_buckets[bucket->next];
            }
            
            assert(bucket->count < BUCKET_ENTRY_COUNT);
            bucket->entries[bucket->count++] = uid;

            inserted = true;
            break;
        }
    }
    rtUnlockMutex(_guard);
    assert(inserted);
    return RT_SUCCESS;
}

rt_result rtAddAssetDependency(rt_uid dependent, rt_uid dependency) {
    rt_result res = InsertIntoList(dependent, dependency, 0);
    if (res != RT_SUCCESS)
        return res;
    res = InsertIntoList(dependency, dependent, 1);
    return res;
}

rt_result rtSaveAssetDependencies(void) {
    assert(rtIsMainThread());
    rt_dependency_file_header header = {.num_lists = 0, .data_size = 0};

    for (size_t i = 0; i < MAP_SIZE; ++i) {
        if (_uids[i] != RT_INVALID_UID) {
            if (!_lists[i].dependencies)
                continue;
            header.num_lists += 1;

            /* Determine the list size */
            rt_dep_list_bucket *bucket = &_buckets[_lists[i].dependencies];
            uint32_t total_list_size   = bucket->count;
            while (bucket->next != END_OF_LIST) {
                bucket = &_buckets[bucket->next];
                total_list_size += bucket->count;
            }

            header.data_size += total_list_size * sizeof(rt_uid) + sizeof(rt_dependency_file_list_header);
        }
    }
    
    FILE *f = fopen("data/deps.bin", "wb");
    if (!f) {
        rtReportError("ASSETC", "Failed to open 'deps.bin' for writing.");
        return RT_UNKNOWN_ERROR;
    }


    if (fwrite(&header, sizeof(header), 1, f) != 1) {
        rtReportError("ASSETC", "Failed to write to 'deps.bin'.");
        fclose(f);
        return RT_UNKNOWN_ERROR;
    }

    void *buffer       = NULL;
    size_t buffer_size = 0;

    for (size_t i = 0; i < MAP_SIZE; ++i) {
        if (_uids[i] != RT_INVALID_UID) {
            if (!_lists[i].dependencies)
                continue;

            /* Determine the list size */
            rt_dep_list_bucket *bucket = &_buckets[_lists[i].dependencies];
            uint32_t total_list_size = bucket->count;
            while (bucket->next != END_OF_LIST) {
                bucket = &_buckets[bucket->next];
                total_list_size += bucket->count;
            }

            /* Allocate */
            size_t required_size =
                total_list_size * sizeof(rt_uid) + sizeof(rt_dependency_file_list_header);
            if (required_size > buffer_size) {
                void *t = realloc(buffer, required_size);
                if (!t) {
                    free(buffer);
                    fclose(f);
                    return RT_UNKNOWN_ERROR;
                }
                buffer = t;
                buffer_size = required_size;
            }

            /* Fill header */
            rt_dependency_file_list_header *list_header = buffer;
            rt_uid *list                                = (rt_uid *)(list_header + 1);
            list_header->uid                            = _uids[i];
            list_header->num_entries                    = total_list_size;

            /* Copy the list */
            uint32_t at = 0;
            bucket = &_buckets[_lists[i].dependencies];
            do {
                memcpy(&list[at], bucket->entries, sizeof(rt_uid) * bucket->count);
                at += bucket->count;
                bucket = &_buckets[bucket->next];
            } while (bucket != &_buckets[END_OF_LIST]);

            XXH64_hash_t hash = XXH3_64bits(list, sizeof(rt_uid) * total_list_size);
            XXH64_canonicalFromHash(&list_header->checksum, hash);

            if (fwrite(buffer, required_size, 1, f) != 1) {
                rtReportError("ASSETC", "Failed to write to 'deps.bin'.");
                fclose(f);
                free(buffer);
                return RT_UNKNOWN_ERROR;
            }
        }
    }

    fclose(f);
    free(buffer);

    return RT_SUCCESS;
}